/*
 *  Subset012.h
 *  LogitReg
 *
 *  Created by Youna Hu on 1/2/11.
 *  Copyright 2011 U of Michigan. All rights reserved.
 *
 */

#ifndef __SUBSET012BYGENE_h__
#define __SUBSET012BYGENE_h__

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <cstdio>
#include <stdlib.h>
#include "StringArray.h"
#include "MathMatrix.h"
#include "Error.h"

///////////////////////////////////////////////////////////////////////////////

class PositionFile {
	
public:
	// Vector of positions
	typedef std::vector<int> PositionVector;
	
	// Chromosome / position pair
	typedef std::vector<PositionVector *> ChromosomeVector;
	
	// Too small
	static const int TOO_SMALL = (-2); // tried to set to be 0
	
	// Too big
	static const int TOO_BIG = (-1);
	
	// Nothing in range
	static const int NOTHING_IN_RANGE = (-3);
	
	// Fail
	static const int SEARCH_FAIL = (-4);
	
	public:
	
	// constructors
	PositionFile(const std::string & strFile);
	PositionFile(const Matrix & posMatrix);
	// destructor
	~PositionFile();
	
	// Find a position in the file
	int FindElement(int iChromosome,int iPosition,bool fEnd);
	
	// Find a range within the file using binary search
	bool FindRange(int iChromosome,int iStart,int iEnd,int & iStartPos,int & iEndPos);
	
private:
	// Chromosomes
	ChromosomeVector m_vecChromosomes;
	
};
	
// define a structure of a gene
struct Gene {
		int iChromosome;
		std::string strName;
		int iStart;
		int iEnd;
	};
	
class GeneFile {
	public:
		// Constructor
		GeneFile(const std::string & strFile);
		
		// Get a gene
		const Gene & GetGene(int iGene) const;
		
		// Number of genes
		size_t GetGeneCount() const;
		
		// get the index of a given gene
		// right now, just use linear search, can upgrade it to binary search later
		int SearchGene(String &geneNameIn) const;
		
	private:
		// Vector of genes
		std::vector<Gene> m_vecGenes;
	};
	
#endif
