#include "LogisticRegression.h"
#include "rareVariantsAnalysis.h"
#include "StringArray.h"
#include "Parameters.h"
#include "StringHash.h"
#include "StringArray.h"
#include "DataFromFiles.h"
#include "Subset012sbyMAF.h"
#include "Subset012byGene.h"	
#include "Table2by2.h"
#include "RVanalysisGlobalPermutation.h"

gsl_rng * g_r;

int main(int argc, char **argv){
		String genofile ("pos.012"); //../vcfReader/subsetGeno.012
    String phenofile("LDL.y.ID");
		String covfile("covFile.ID.2.txt");
		String geneList("outGeneSorted.txt");	
		String globalPermuteSummaryFile("globalPermuteSummary.txt");
		String outPfile("geneTestPvalues.txt");
		String weightFile("");	
		String snpTestFile("snpTestPvalues.txt");
		String collapseChoice("so");
	
    // command line input set up
    ParameterList pl;
    int nPermuteIn = 1;
    double mafCut = 0.01;
    int rand_seed = 1;
		int covConsider = 0;
	
    // command line specification
    BEGIN_LONG_PARAMETERS(options)
      LONG_PARAMETER_GROUP("GENOTYPE")
      LONG_STRINGPARAMETER("genofile", &genofile)
	    LONG_STRINGPARAMETER("geneList", &geneList)
	    LONG_DOUBLEPARAMETER("cutoff",&mafCut)
			LONG_STRINGPARAMETER("collapseChoice",&collapseChoice) //(or,sum,wt)//,user)
			//LONG_STRINGPARAMETER("userInputWeights",&weightFile)
      LONG_PARAMETER_GROUP("PHENOTYPE")
      LONG_STRINGPARAMETER("phenofile", &phenofile)
	    LONG_PARAMETER_GROUP("COVARIATES")
			LONG_INTPARAMETER("covConsider",&covConsider) //(Yes:1,No:0)
			LONG_STRINGPARAMETER("covfile", &covfile)
      LONG_PARAMETER_GROUP("PERMUTATION")
      LONG_INTPARAMETER("nPermute", &nPermuteIn)
      LONG_INTPARAMETER("PermutationSeed", &rand_seed)
			LONG_PARAMETER_GROUP("GENE LEVEL TEST RESULT")
    	LONG_STRINGPARAMETER("geneGlobalTestOut",&globalPermuteSummaryFile)
			LONG_STRINGPARAMETER("geneTestpvalueFile",&outPfile)
    END_LONG_PARAMETERS();

    pl.Add((new LongParameters("RARE VARIANT ANALYSIS OPTIONS:",options))->SetPrecision(3));
    pl.Read(argc,argv);
    pl.Status();
	
	// read the file into RAM and manipultate it
	Matrix Xbig;
	Vector Ybig;
	Matrix COVbig;
	String indvFileName(genofile + ".indv");
	String posFileName(genofile+".pos");
	std::string geneListFileName(geneList);
	Matrix positionMatrix;
	String frqInputFileName(genofile+".frq");
	std::string collapseChoiceIn = collapseChoice.c_str();

	// Whether to update userWeight or not
	Vector userInputWeights(1); // if user hasn't changed the weightFile parameter, we just specify a simple vector
	if (weightFile != "") { // if user has specified a file for the weights, then read in this weight function
		FILE * inWtFile = fopen(weightFile, "r");
		if (inWtFile == NULL) {
			error("Error in opening %s !\n", weightFile.c_str());
		} else {
			printf("Opening file %s ...\n",weightFile.c_str());
		}

	}
	
	DataFromFiles dataFile(false,true);
	// read in the genotype matrix
	printf("Read in genotypes from file (%s)......\n",genofile.c_str());
	dataFile.File2Matrix(genofile, Xbig);
	printf("Done.\n");
	if(!Subset012sbyMAF::CheckGenoValid(Xbig)){
		error("Genotype should be either 0, 1 or 2.\n");
	}
	// read in the phenotype vector
	printf("Read in phenotype from file (%s) ......\n",phenofile.c_str());
	dataFile.File2Vector(phenofile, indvFileName, Ybig);
	printf("Done.\n");
	if (!Subset012sbyMAF::CheckPhenoValid(Ybig)){
		error("Phenotype should be either 0 or 1. \n");
	}
	// read in the covariate matrixs
	if (covConsider == 1) {
		printf("Read in Covariate matrix from file (%s) ......\n",covfile.c_str());
		dataFile.File2Matrix(covfile, indvFileName, COVbig);
		printf("Done.\n");
	}

	printf("genotype matrix size (%d,%d)\n",Xbig.rows, Xbig.cols);
	printf("phenotype size (%d)\n",Ybig.Length());
	printf("covariate matrix size (%d, %d)\n", COVbig.rows, COVbig.cols);
	int nSampleX = Xbig.rows, nSampleY = Ybig.Length(), nSampleCOV = COVbig.rows;
	
	// sanity check for input data
	if (covConsider == 1) {
		if (nSampleX != nSampleY || nSampleX != nSampleCOV){
			printf("Error: Invalid input files.\n");
			error("Different sample sizes: (genotype file: n = %d, phenotype file: n = %d, covariate file: n = %d).\n",
						 nSampleX, nSampleY, nSampleCOV);
		}		
	} else {
		if (nSampleX != nSampleY) {
			printf("Error: Invalid input files.\n");
			error("Different sample sizes: (genotype file: n = %d, phenotype file: n = %d).\n",
						 nSampleX, nSampleY, nSampleCOV);
		}
	}

	// read in position file and sorted gene list file
	// read the positionFile into the positionMatrix 
	// note: need to pay attention when chrosome X and Y are considered
	dataFile.PositionFile2Matrix(posFileName, positionMatrix); 
	
	// Now that the genotype matrix and the position file are read into matrices
	// We can subset both of them so as to only select rare variants (Xbig,positionMatrix)
	printf("Now trim the matrix by mafCut = %f...\n",mafCut);
	Subset012sbyMAF::GenoPosSubsetbyMAF(positionMatrix, Xbig, frqInputFileName, mafCut);
	printf("The trimmed genotype matrix size (%d,%d)\n",Xbig.rows,Xbig.cols);
	
	// make sure that the reference allele is the major allele
	printf("Now flip the reference allele to to be the major allele.....\n");
	Subset012sbyMAF::FlipGenoReference(Xbig);
	printf("Done.\n");

	/*************** Do rare variant analysis *******************/
	// Get the asympototic p value for each gene
	// Do global permutation (get the p value from permutation, get the 5% and 95% quantile across genes at each permutation)

	//************ initialize the gsl random number starter *************//
	gsl_rng_env_setup();
	g_r = gsl_rng_alloc(gsl_rng_default);
  gsl_rng_set(g_r,rand_seed); 
	//************ done inditialize the gsl random number starter	**************//
	
	if (covConsider == 1) { // when consider covariate, use permutation
		
		RareVariantAnalysis RVtest(Xbig,COVbig,Ybig,collapseChoiceIn,1); 
		RVanalysisGlobalPermutation testPermute(RVtest,Xbig,Ybig,nPermuteIn,positionMatrix,geneListFileName);
		
		printf("Conducting rare variant test with covariates (using %d permuations) ......\n", nPermuteIn);
		testPermute.PvaluesGenePermutationWithCov(Xbig, COVbig, Ybig,collapseChoiceIn,userInputWeights);
		printf("Done.\n");
		
		printf("Write p values into file (%s) ...... ",outPfile.c_str());
		testPermute.WritePs(outPfile,true); 
		printf("Done.\n");
		
		printf("Write the 5%% and 95%% quantile p values for each permutation into file (%s)......\n",globalPermuteSummaryFile.c_str());
		testPermute.WriteGlobalPermuteSummary(globalPermuteSummaryFile);		
		printf("Done.\n");
	} else { // when we don't consider covariates, use asymptotic distribution, 
					 // since I have done comparisons to verify that it works
		RareVariantAnalysis RVtest(Xbig,Ybig,collapseChoiceIn, nPermuteIn);
		RVanalysisGlobalPermutation testPermute(RVtest,Xbig,Ybig,nPermuteIn,positionMatrix,geneListFileName);
		
		// if the number of permutation is less than 10, then use asymptotic distribution
		if (nPermuteIn < 10) {
			
			printf("Conducting rare variant test without covariates (using asymptotic distribution) ......\n");
			testPermute.PvaluesGeneAsy(Xbig,Ybig,collapseChoiceIn,userInputWeights);
			printf("Done.\n");
			
			printf("Write p values into file %s ......\n", outPfile.c_str());
			testPermute.WritePs(outPfile, false); // output the asymptotic p values
			printf("Done.\n");
			
		} else {
			printf("Conducting rare variant test with covariates (using %d permuations) ......\n", nPermuteIn);
			testPermute.PvaluesGenePermutation(Xbig, Ybig, collapseChoiceIn, userInputWeights);
			printf("Done.\n");
			
			printf("Write p values into file %s ......\n", outPfile.c_str());
			testPermute.WritePs(outPfile,true); 
			printf("Done.\n");
			
			printf("Write the 5%% and 95%\% quantile p values for each permutation into file (%s)......\n",globalPermuteSummaryFile.c_str());
			testPermute.WriteGlobalPermuteSummary(globalPermuteSummaryFile);	
			printf("Done.\n");
		}
	}
	
	// ******* delete the random number starter ********** //
	gsl_rng_free(g_r);
	// ***** done deleting the random number starter *********//
	
	return 1;
}

